@@ -34,23 +34,17 @@ lpcde_fn = function(y_data, x_data, y_grid, x, p, q, p_RBC, q_RBC, bw, mu, nu,
                    mu=mu, nu=nu, h=bw, kernel_type=kernel_type)
   est = f_hat_val$est
   eff.n = f_hat_val$eff.n
+  est_flag = f_hat_val$singular_flag
 
   # covariance matrix
 
   # standard errors
   # if(var_type=="ustat"){
-    covMat = cov_hat(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid, p=p, q=q,
+  covmat = cov_hat(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid, p=p, q=q,
                    mu=mu, nu=nu, h=bw, kernel_type=kernel_type)
-    se = sqrt(abs(diag(covMat)))
-  # }else if (var_type == "simul"){
-  #   covMat = simul_cov(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid, p=p, q=q,
-  #                  mu=mu, nu=nu, h=bw, kernel_type=kernel_type)
-  #   se = sqrt(abs(diag(covMat$covMat)))
-  # }else if (var_type == "asymp"){
-    # covMat = asymp_var(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid, p=p, q=q,
-                       # mu=mu, nu=nu, h=bw, kernel_type=kernel_type)
-    # se = sqrt(abs(diag(covMat)))
-  # }
+  covMat = covmat$cov
+  c_flag = covmat$singular_flag
+  se = sqrt(abs(diag(covMat)))
 
   if (rbc){
     est_rbc = matrix(0L, nrow = length(y_grid), ncol = 1)
@@ -59,38 +53,41 @@ lpcde_fn = function(y_data, x_data, y_grid, x, p, q, p_RBC, q_RBC, bw, mu, nu,
       est_rbc = est
       se_rbc = se
       covMat_rbc = covMat
+
+      rbc_flag = f_hat_val$singular_flag
+      c_rbc_flag = covmat$singular_flag
     }else{
       # estimate
-      est_rbc = fhat(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid, p=p_RBC,
-                     q=q_RBC, mu=mu, nu=nu, h=bw, kernel_type=kernel_type)$est
+      f_hat_rbc = fhat(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid, p=p_RBC,
+                    q=q_RBC, mu=mu, nu=nu, h=bw, kernel_type=kernel_type)
 
+      est_rbc =  f_hat_rbc$est
+      rbc_flag = f_hat_rbc$singular_flag
       # covariance matrix
 
       # standard errors
       # if(var_type=="ustat"){
-        covMat_rbc = cov_hat(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid,
-                           p=p_RBC, q=q_RBC, mu=mu, nu=nu, h=bw,
-                           kernel_type=kernel_type)
-        se_rbc = sqrt(abs(diag(covMat_rbc)))
-      # }else if (var_type == "simul"){
-      #   covMat_rbc = simul_cov(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid, p=p_RBC, q=q_RBC,
-      #                  mu=mu, nu=nu, h=bw, kernel_type=kernel_type)
-      #   se_rbc = sqrt(abs(diag(covMat$covMat)))
-      # }else if (var_type == "asymp"){
-        # covMat_rbc = asymp_var(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid, p=p_RBC, q=q_RBC,
-                           # mu=mu, nu=nu, h=bw, kernel_type=kernel_type)
-        # se_rbc = sqrt(abs(diag(covMat_rbc)))
-      # }
+      covmat_rbc = cov_hat(x_data=x_data, y_data=y_data, x=x, y_grid=y_grid,
+                       p=p_RBC, q=q_RBC, mu=mu, nu=nu, h=bw,
+                       kernel_type=kernel_type)
+      covMat_rbc = covmat_rbc$cov
+      c_rbc_flag = covmat_rbc$singular_flag
+      se_rbc = sqrt(abs(diag(covMat_rbc)))
     }
 
     # with rbc results
+    if (est_flag==TRUE | c_flag==TRUE | rbc_flag==TRUE | c_rbc_flag==TRUE){
+      singular_flag = TRUE
+    }else{
+      singular_flag=FALSE
+    }
     estimate = cbind(y_grid, bw, est, est_rbc, se, se_rbc)
     colnames(estimate) = c("y_grid","bw", "est","est_RBC", "se", "se_RBC")
     rownames(estimate) = c()
     est_result = list("est" = estimate,
                       "CovMat" = list("CovMat" = covMat,
                                       "CovMat_RBC" = covMat_rbc),
-                      "x" = x, "eff_n" = eff.n)
+                      "x" = x, "eff_n" = eff.n, "singular_flag" = singular_flag)
   } else{
     # without rbc results
     # setting rbc values to be the same as non-rbc values
@@ -104,7 +101,7 @@ lpcde_fn = function(y_data, x_data, y_grid, x, p, q, p_RBC, q_RBC, bw, mu, nu,
     est_result = list("est" = estimate,
                       "CovMat" = list("CovMat" = covMat,
                                       "CovMat_RBC" = covMat_rbc),
-                      "x" = x, "eff_n" = eff.n)
+                      "x" = x, "eff_n" = eff.n, "singular_flag" = singular_flag)
   }
 
   return(est_result)
@@ -129,6 +126,8 @@ fhat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
   n = length(y_data)
   d = ncol(x_data)
   ng = length(y_grid)
+  x = matrix(x, ncol=d)
+  singular_flag=FALSE
 
   # x basis vector
   e_nu = basis_vec(x, q, nu)
@@ -142,9 +141,9 @@ fhat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
   if (length(unique(h))==1){
     h = h[1]
     # localization for x
-    idx = which(abs(x_data-x)<=h^d)
+    idx = which(rowSums(abs(sweep(x_data, 2, x))<=h)==d)
 
-    x_idx = x_data[idx]
+    x_idx = matrix(x_data[idx, ], ncol=d)
     y_idx = y_data[idx]
 
     # idx of ordering wrt y
@@ -152,11 +151,16 @@ fhat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
 
     # sorting datasets
     y_sorted = as.matrix(y_idx[sort_idx])
-    x_sorted = as.matrix(x_idx[sort_idx])
+    x_sorted = matrix(x_idx[sort_idx, ], ncol=d)
 
     # x constants
-    x_scaled = (x_sorted-x)/(h^d)
-    sx_mat = solve(S_x(x_scaled, q, kernel_type)/(n*h^d))
+    x_scaled = sweep(x_sorted, 2, x)/(h^d)
+    if(check_inv(S_x(x_scaled, q, kernel_type)/(n*h^d))[1]== TRUE){
+      sx_mat = solve(S_x(x_scaled, q, kernel_type)/(n*h^d))
+    } else{
+      singular_flag = TRUE
+      sx_mat = matrix(0L, nrow = length(e_nu), ncol = length(e_nu))
+    }
     bx = b_x(x_scaled, sx_mat, e_nu, q, kernel_type)
 
     for (j in 1:ng){
@@ -169,10 +173,20 @@ fhat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
         nh_vec [j] = length(y_elems)
       } else {
         if (mu==0){
-          sx_mat = solve(S_x(as.matrix(x_scaled[y_elems]), q, kernel_type)/(n*h^d))
+          if(check_inv(S_x(as.matrix(x_scaled[y_elems, ]), q, kernel_type)/(n*h^d))[1]== TRUE){
+            sx_mat = solve(S_x(as.matrix(x_scaled[y_elems, ]), q, kernel_type)/(n*h^d))
+          } else{
+            singular_flag = TRUE
+            sx_mat = matrix(0L, nrow = length(e_nu), ncol = length(e_nu))
+          }
           bx = b_x(x_scaled, sx_mat, e_nu, q, kernel_type)
           # sy matrix
-          sy_mat = solve(S_x(as.matrix(y_scaled), p, kernel_type)/(n*h))
+          if(check_inv(S_x(as.matrix(y_scaled), p, kernel_type)/(n*h))[1]== TRUE){
+            sy_mat = solve(S_x(as.matrix(y_scaled), p, kernel_type)/(n*h))
+          } else{
+            singular_flag = TRUE
+            sy_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+          }
 
           # y constants
           ax = b_x(y_scaled, sy_mat, e_mu, p, kernel_type)
@@ -184,10 +198,15 @@ fhat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
           nh_vec[j] = length(y_elems)
         } else {
           # sy matrix
-          sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))
+          if(check_inv(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))[1]== TRUE){
+            sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))
+          } else{
+            singular_flag = TRUE
+            sy_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+          }
 
           # y constants
-          ax = b_x(y_scaled[y_elems], sy_mat, e_mu, p, kernel_type)
+          ax = b_x(as.matrix(y_scaled[y_elems]), sy_mat, e_mu, p, kernel_type)
 
           # adding and multiplying
           f_hat[j] = ax %*% cumsum(bx[y_elems])
@@ -203,197 +222,89 @@ fhat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
   } else {
     for (j in 1:ng){
       # localization for x
-      idx = which(abs(x_data-x)<=h[j])
+      idx = which(rowSums(abs(sweep(x_data, 2, x))<=h[j])==d)
 
-      x_data_loc = x_data[idx]
+      x_data_loc = matrix(x_data[idx, ], ncol=d)
       y_data_loc = y_data[idx]
 
       # idx of ordering wrt y
       sort_idx = sort(y_data_loc, index.return=TRUE)$ix
 
       # sorting datasets
-      y_data_loc = as.matrix(y_data_loc[sort_idx])
-      x_data_loc = as.matrix(x_data_loc[sort_idx])
+      y_data_loc = matrix(y_data_loc[sort_idx])
+      x_data_loc = matrix(x_data_loc[sort_idx, ], ncol=d)
 
       # x constants
-      x_scaled = (x_data_loc-x)/(h[j]^d)
-      sx_mat = solve(S_x(as.matrix(x_scaled), q, kernel_type)/(n*h[j]^d))
+      x_scaled = sweep(x_data_loc, 2, x)/(h[j]^d)
+      if(check_inv(S_x(as.matrix(x_scaled/(n*h[j]^d)), q, kernel_type))[1]== TRUE){
+        sx_mat = solve(S_x(as.matrix(x_scaled/(n*h[j]^d)), q, kernel_type))
+      } else{
+        singular_flag = TRUE
+        sx_mat = matrix(0L, nrow = length(e_nu), ncol = length(e_nu))
+      }
       bx = b_x(as.matrix(x_scaled), sx_mat, e_nu, q, kernel_type)
 
       y = y_grid[j]
       y_scaled = (y_data_loc-y)/h[j]
       y_elems = which(abs(y_scaled)<=1)
 
-      # sy matrix
-      sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[j]))
+      if (length(y_elems)<=5){
+        f_hat[j] = 0
+        nh_vec [j] = length(y_elems)
+      } else {
+        if (mu==0){
+          if(check_inv(S_x(as.matrix(x_scaled[y_elems, ]/(n*h[j]^d)), q, kernel_type))[1]== TRUE){
+            sx_mat = solve(S_x(S_x(as.matrix(x_scaled[y_elems, ]/(n*h[j]^d)), q, kernel_type)))
+          } else{
+            singular_flag = TRUE
+            sx_mat = matrix(0L, nrow = length(e_nu), ncol = length(e_nu))
+          }
+          bx = b_x(as.matrix(x_scaled), sx_mat, e_nu, q, kernel_type)
+          # sy matrix
+          if(check_inv(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[j]))[1]== TRUE){
+            sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[j]))
+          } else{
+            singular_flag = TRUE
+            sy_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+          }
 
-      # y constants
-      ax = b_x(y_scaled[y_elems], sy_mat, e_mu, p, kernel_type)
+         # y constants
+          ax = b_x(matrix(y_scaled[y_elems]), sy_mat, e_mu, p, kernel_type)
 
-      # adding and multiplying
-      f_hat[j] = ax %*% cumsum(bx[y_elems])
+         # adding and multiplying
+          f_hat[j] = ax %*% cumsum(bx[y_elems])
 
-      # scaling
-      f_hat[j] = f_hat[j]/(n^2*h[j]^(d+mu+nu+1))
-      # number of datapoints used
-      nh_vec[j] = length(y_elems)
+         # scaling
+          f_hat[j] = f_hat[j]/(n^2*h[j]^(d+mu+nu+1))
+         # number of datapoints used
+          nh_vec[j] = length(y_elems)
+        } else{
+         # sy matrix
+          if(check_inv(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[j]))[1]== TRUE){
+            sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[j]))
+          } else{
+            singular_flag = TRUE
+            sy_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+          }
+
+         # y constants
+          ax = b_x(matrix(y_scaled[y_elems]), sy_mat, e_mu, p, kernel_type)
+
+         # adding and multiplying
+          f_hat[j] = ax %*% cumsum(bx[y_elems])
+
+         # scaling
+          f_hat[j] = f_hat[j]/(n^2*h[j]^(d+mu+nu+1))
+         # number of datapoints used
+          nh_vec[j] = length(y_elems)
+        }
+      }
     }
   }
-  return(list("est" = f_hat, "eff.n" = nh_vec))
+  return(list("est" = f_hat, "eff.n" = nh_vec, "singular_flag" = singular_flag))
 
 }
 
-##' @title asymp_var: asymptotic covariance estimator
-##' @description Function for estimating the variance-covariance matrix.
-##' @param y_data response variable dataset, vector.
-##' @param x_data covariate dataset, vector or matrix.
-##' @param y_grid Numeric vector, specifies the grid of evaluation points along y-direction.
-##' @param x Numeric vector or matrix, specifies the grid of evaluation points along x-direction.
-##' @param p polynomial order for y.
-##' @param q polynomial order for covariates.
-##' @param h Numeric, bandwidth vector.
-##' @param mu degree of derivative with respect to y.
-##' @param nu degree of derivative with respect to x.
-##' @param kernel_type kernel function choice.
-##' @return covariance matrix for all the grid points
-##' @keywords internal
-#asymp_var = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
-
-#  n = length(y_data)
-#  d = ncol(x_data)
-#  ng = length(y_grid)
-
-#  # x basis vector
-#  e_nu = basis_vec(x, q, nu)
-
-#  # y basis vector
-#  e_mu = basis_vec(1, p, mu)
-
-#  if(length(unique(h))==1){
-#    h = h[1]
-#    theta = fhat(x_data=as.matrix(x_data), y_data=as.matrix(y_data), x=x,
-#                 y_grid=y_grid, p=2, q=1, mu=1, nu=0, h=h,
-#                 kernel_type=kernel_type)$est
-#    theta_00 = fhat(x_data=as.matrix(x_data), y_data=as.matrix(y_data), x=x,
-#                    y_grid=y_grid, p=1, q=1, mu=0, nu=0, h=h,
-#                    kernel_type=kernel_type)$est
-
-#    # localization for x
-#    idx = which(abs(x_data-x)<=h)
-
-#    x_data = x_data[idx]
-#    y_data = y_data[idx]
-
-#    # idx of ordering wrt y
-#    sort_idx = sort(y_data, index.return=TRUE)$ix
-
-#    # sorting datasets
-#    y_data = as.matrix(y_data[sort_idx])
-#    x_data = as.matrix(x_data[sort_idx])
-
-#    # x constants
-#    x_scaled = (x_data-x)/(h^d)
-
-#    sx_mat = solve(S_x(as.matrix(x_scaled), q, kernel_type)/(n*h^d))
-#    Tx = T_x(x_scaled, x, q, h, kernel_type)/h^d
-
-#    c_mat = matrix(0L, nrow=ng, ncol=ng)
-#    for (j in 1:ng){
-#      for (k in 1:j){
-#        y = y_grid[j]
-#        yp = y_grid[k]
-#        y_scaled = (y_data-y)/h
-#        yp_scaled = (y_data-yp)/h
-#        y_elems = which(abs(y_scaled)<1)
-#        yp_elems = which(abs(yp_scaled)<1)
-#        sy_mat = solve(S_x(as.matrix(y_scaled), p, kernel_type)/(n*h))
-#        syp_mat = solve(S_x(as.matrix(yp_scaled), p, kernel_type)/(n*h))
-#        elems = intersect(y_elems, yp_elems)
-#        if (length(elems)==0){
-#          c_mat[j, k] = 0
-#          c_mat[k, j] = c_mat[j, k]
-#        }else{
-#          if (mu==0){
-#            c_mat[j, k] = theta_00[j]*(1-theta_00[j]) * (sx_mat%*%Tx%*%sx_mat)[nu+1, nu+1]
-#            c_mat[k, j] = c_mat[j, k]
-#          }else{
-#            Ty = T_y(y_scaled[elems], yp_scaled[elems], p, kernel_type)/(choose(n, 2)*h^2)
-#            c_mat[j, k] = theta[j] * (sy_mat%*%Ty%*%syp_mat)[mu+1, mu+1] * (sx_mat%*%Tx%*%sx_mat)[nu+1, nu+1]
-#            c_mat[k, j] = c_mat[j, k]
-#          }
-#        }
-#      }
-#    }
-#    if(mu == 0){
-#      c_mat = sweep(sweep(c_mat, MARGIN=1, FUN="*", STATS=1/(n*h^(d+2*nu))),
-#                                            MARGIN=2, FUN="*", STATS=1/(n*h^(d+2*nu)))
-#    }else{
-#      c_mat = sweep(sweep(c_mat, MARGIN=1, FUN="*", STATS=1/(n*h^(d+2*mu+2*nu-1))),
-#                    MARGIN=2, FUN="*", STATS=1/(n*h^(d+2*mu+2*nu-1)))
-#    }
-#  }else{
-
-#    theta = fhat(x_data=as.matrix(x_data), y_data=as.matrix(y_data), x=x, y_grid=y_grid, p=2,
-#                      q=1, mu=1, nu=0, h=h, kernel_type=kernel_type)$est
-#    theta_00 = fhat(x_data=as.matrix(x_data), y_data=as.matrix(y_data), x=x, y_grid=y_grid, p=3,
-#                            q=1, mu=0, nu=0, h=h, kernel_type=kernel_type)$est
-#    # localization for x
-#    idx = which(abs(x_data-x)<=mean(h))
-
-#    x_data = x_data[idx]
-#    y_data = y_data[idx]
-
-#    # idx of ordering wrt y
-#    sort_idx = sort(y_data, index.return=TRUE)$ix
-
-#    # sorting datasets
-#    y_data = as.matrix(y_data[sort_idx])
-#    x_data = as.matrix(x_data[sort_idx])
-
-#    # x constants
-#    x_scaled = (x_data-x)/(mean(h)^d)
-
-#    sx_mat = solve(S_x(as.matrix(x_scaled), q, kernel_type)/(n*mean(h)^d))
-#    Tx = T_x(x_scaled, x, q, mean(h), kernel_type)/(mean(h)^d)
-
-#    c_mat = matrix(0L, nrow=ng, ncol=ng)
-#    for (j in 1:ng){
-#      for (k in 1:j){
-#        y = y_grid[j]
-#        yp = y_grid[k]
-#        y_scaled = (y_data-y)/h[j]
-#        yp_scaled = (y_data-yp)/h[k]
-#        y_elems = which(abs(y_scaled)<1)
-#        yp_elems = which(abs(yp_scaled)<1)
-#        sy_mat = solve(S_x(as.matrix(y_scaled), p, kernel_type)/(n*h[j]))
-#        syp_mat = solve(S_x(as.matrix(yp_scaled), p, kernel_type)/(n*h[k]))
-#        elems = intersect(y_elems, yp_elems)
-#        if (length(elems)==0){
-#          c_mat[j, k] = 0
-#          c_mat[k, j] = c_mat[j, k]
-#        }else {
-#          if (mu==0){
-#            c_mat[j, k] = theta_00[j]*(1-theta_00[j]) * (sx_mat%*%Tx%*%sx_mat)[nu+1, nu+1]
-#            c_mat[k, j] = c_mat[j, k]
-#          }else{
-#            Ty = T_y(y_scaled[elems], yp_scaled[elems], p, kernel_type)/(choose(n, 2)*h[j]*h[k])
-#            c_mat[j, k] = theta[j] * (sy_mat%*%Ty%*%syp_mat)[mu+1, mu+1] * (sx_mat%*%Tx%*%sx_mat)[nu+1, nu+1]
-#            c_mat[k, j] = c_mat[j, k]
-#          }
-#        }
-#      }
-#    }
-#    if(mu == 0){
-#      c_mat = sweep(sweep(c_mat, MARGIN=1, FUN="*", STATS=1/(n*h^(d+2*nu))),
-#                                            MARGIN=2, FUN="*", STATS=1/(n*h^(d+2*nu)))
-#    }else{
-#      c_mat = sweep(sweep(c_mat, MARGIN=1, FUN="*", STATS=1/(n*h^(d+2*mu+2*nu-1))),
-#                    MARGIN=2, FUN="*", STATS=1/(n*h^(d+2*mu+2*nu-1)))
-#    }
-#  }
-#  return(c_mat)
-#}
-
 #' @title cov_hat: covariance estimator
 #' @description Function for estimating the variance-covariance matrix.
 #' @param y_data response variable dataset, vector.
@@ -416,6 +327,7 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
   d = ncol(x_data)
   ng = length(y_grid)
 
+  singular_flag = FALSE
   # x basis vector
   e_nu = basis_vec(x, q, nu)
 
@@ -425,21 +337,26 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
   if(length(unique(h))==1){
     h = h[1]
     # localization for x
-    idx = which(abs(x_data-x)<=h)
+    idx = which(rowSums(abs(sweep(x_data, 2, x))<=h)==d)
 
-    x_data = x_data[idx]
-    y_data = y_data[idx]
+    x_idx = as.matrix(x_data[idx, ])
+    y_idx = y_data[idx]
 
     # idx of ordering wrt y
-    sort_idx = sort(y_data, index.return=TRUE)$ix
+    sort_idx = sort(y_idx, index.return=TRUE)$ix
 
     # sorting datasets
-    y_data = as.matrix(y_data[sort_idx])
-    x_data = as.matrix(x_data[sort_idx])
+    y_sorted = as.matrix(y_idx[sort_idx])
+    x_sorted = as.matrix(x_idx[sort_idx, ])
 
     # x constants
-    x_scaled = (x_data-x)/(h^d)
-    sx_mat = solve(S_x(as.matrix(x_scaled), q, kernel_type)/(n*h^d))
+    x_scaled = sweep(x_sorted, 2, x)/(h^d)
+    if(check_inv(S_x(x_scaled, q, kernel_type)/(n*h^d))[1]== TRUE){
+      sx_mat = solve(S_x(x_scaled, q, kernel_type)/(n*h^d))
+    } else{
+      singular_flag = TRUE
+      sx_mat = matrix(0L, nrow = length(e_nu), ncol = length(e_nu))
+    }
     bx = b_x(as.matrix(x_scaled), sx_mat, e_nu, q, kernel_type)
 
     # initialize matrix
@@ -450,25 +367,41 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
         # relevant entries wrt y and y_prime
         y = y_grid[i]
         y_prime = y_grid[j]
-        y_scaled = (y_data - y)/h
-        yp_scaled = (y_data-y_prime)/h
+        y_scaled = (y_sorted - y)/h
+        yp_scaled = (y_sorted-y_prime)/h
         y_elems = which(abs(y_scaled)<=1)
         yp_elems = which(abs(yp_scaled)<=1)
         elems = intersect(y_elems, yp_elems)
 
         if ( length(elems) <= 5){
-          c_hat[i, j] = c_hat[i, j]
+          c_hat[i, j] = 0
+          c_hat[j, i] = 0
         } else{
           if (mu==0){
-            sx_mat = solve(S_x(as.matrix(x_scaled[y_elems]), q, kernel_type)/(n*h^d))
+            if(check_inv(S_x(as.matrix(x_scaled[y_elems, ]/(n*h^d)), q, kernel_type))[1]== TRUE){
+              sx_mat = solve(S_x(as.matrix(x_scaled[y_elems, ]/(n*h^d)), q, kernel_type))
+            } else{
+              singular_flag = TRUE
+              sx_mat = matrix(0L, nrow = length(e_nu), ncol = length(e_nu))
+            }
             bx = b_x(as.matrix(x_scaled), sx_mat, e_nu, q, kernel_type)
             # sy matrix
-            sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))
-            syp_mat = solve(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h))
+            if(check_inv(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))[1]== TRUE){
+              sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))
+            } else{
+              singular_flag = TRUE
+              sy_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+            }
+            if(check_inv(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h))[1]== TRUE){
+              syp_mat = solve(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h))
+            } else{
+              singular_flag = TRUE
+              syp_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+            }
 
             # computing y and yprime vectors
-            a_y = b_x(y_scaled[elems], sy_mat, e_mu, p, kernel_type)
-            a_yp =  b_x(yp_scaled[elems], syp_mat, e_mu, p, kernel_type)
+            a_y = b_x(matrix(y_scaled[elems]), sy_mat, e_mu, p, kernel_type)
+            a_yp =  b_x(matrix(yp_scaled[elems]), syp_mat, e_mu, p, kernel_type)
 
             # part 1 of the sum
             aj = cumsum(a_y)
@@ -476,12 +409,22 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
 
           }else{
             # sy matrix
-            sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))
-            syp_mat = solve(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h))
+            if(check_inv(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))[1]== TRUE){
+              sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h))
+            } else{
+              singular_flag = TRUE
+              sy_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+            }
+            if(check_inv(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h))[1]== TRUE){
+              syp_mat = solve(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h))
+            } else{
+              singular_flag = TRUE
+              syp_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+            }
 
             # computing y and yprime vectors
-            a_y = b_x(y_scaled[elems], sy_mat, e_mu, p, kernel_type)
-            a_yp =  b_x(yp_scaled[elems], syp_mat, e_mu, p, kernel_type)
+            a_y = b_x(as.matrix(y_scaled[y_elems]), sy_mat, e_mu, p, kernel_type)
+            a_yp =  b_x(as.matrix(yp_scaled[elems]), syp_mat, e_mu, p, kernel_type)
 
             # part 1 of the sum
             aj = cumsum(a_y)
@@ -494,7 +437,7 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
             t_2 = (n-k) * a_yp[k] * aj[k]
             t_3 = (n-k) * a_y[k] * ak[k]
             t_4 = (n-k)^2 * a_y[k] * a_yp[k]
-            c_hat[i, j] = c_hat[i, j] + bx[elems[k]]^2 * (t_1 + t_2 + t_3 + t_4)
+            c_hat[i, j] = c_hat[i, j] + bx[1, elems[k]]^2 * (t_1 + t_2 + t_3 + t_4)
           }
         }
         # estimated means
@@ -515,7 +458,6 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
         c_hat[i, j] = c_hat[i, j]/(n*(n-1)^2) - theta_y*theta_yp/n^2
         c_hat[j, i] = c_hat[i, j]
 
-
       }
     }
     if(mu==0){
@@ -533,23 +475,29 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
 
     for (i in 1:ng){
       for (j in 1:i){
+        hx = mean(h[i], h[j])
         # localization for x
-        idx = which(abs(x_data-x)<=h[i])
+        idx = which(rowSums(abs(sweep(x_data, 2, x))<=hx)==d)
 
-        x_data_loc = x_data[idx]
+        x_data_loc = matrix(x_data[idx, ], ncol=d)
         y_data_loc = y_data[idx]
 
         # idx of ordering wrt y
         sort_idx = sort(y_data_loc, index.return=TRUE)$ix
 
         # sorting datasets
-        y_data_loc = as.matrix(y_data_loc[sort_idx])
-        x_data_loc = as.matrix(x_data_loc[sort_idx])
+        y_data_loc = matrix(y_data_loc[sort_idx])
+        x_data_loc = matrix(x_data_loc[sort_idx, ], ncol=d)
 
         # x constants
-        x_scaled = (x_data_loc-x)/(h[i]^d)
-        sx_mat = solve(S_x(x_scaled, q, kernel_type)/(n*h[i]^d))
-        bx = b_x(x_scaled, sx_mat, e_nu, q, kernel_type)
+        x_scaled = sweep(x_data_loc, 2, x)/(hx^d)
+        if(check_inv(S_x(matrix(x_scaled/(n*hx^d), ncol=d), q, kernel_type))[1]== TRUE){
+          sx_mat = solve(S_x(matrix(x_scaled/(n*hx^d), ncol=d), q, kernel_type))
+        } else{
+          singular_flag = TRUE
+          sx_mat = matrix(0L, nrow = length(e_nu), ncol = length(e_nu))
+        }
+        bx = b_x(matrix(x_scaled, ncol=d), sx_mat, e_nu, q, kernel_type)
 
         # relevant entries wrt y and y_prime
         y = y_grid[i]
@@ -560,19 +508,35 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
         yp_elems = which(abs(yp_scaled)<=1)
         elems = intersect(y_elems, yp_elems)
 
-        if ( length(elems) == 0){
-          c_hat[i, j] = c_hat[i, j]
+        if ( length(elems) <= 5){
+          c_hat[i, j] = 0
+          c_hat[j, i] = 0
         } else{
           if (mu==0){
-            sx_mat = solve(S_x(as.matrix(x_scaled[y_elems]), q, kernel_type)/(n*h[i]^d))
+            if(check_inv(S_x(as.matrix(x_scaled[y_elems, ]/(n*hx^d)), q, kernel_type))[1]== TRUE){
+              sx_mat = solve(S_x(as.matrix(x_scaled[y_elems, ]/(n*hx^d)), q, kernel_type))
+            } else{
+              singular_flag = TRUE
+              sx_mat = matrix(0L, nrow = length(e_nu), ncol = length(e_nu))
+            }
             bx = b_x(as.matrix(x_scaled), sx_mat, e_nu, q, kernel_type)
             # sy matrix
-            sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[i]))
-            syp_mat = solve(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h[j]))
+            if(check_inv(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[i]))[1]== TRUE){
+              sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[i]))
+            } else{
+              singular_flag = TRUE
+              sy_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+            }
+            if(check_inv(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h[j]))[1]== TRUE){
+              syp_mat = solve(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h[j]))
+            } else{
+              singular_flag = TRUE
+              syp_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+            }
 
             # computing y and yprime vectors
-            a_y = b_x(y_scaled[elems], sy_mat, e_mu, p, kernel_type)
-            a_yp =  b_x(yp_scaled[elems], syp_mat, e_mu, p, kernel_type)
+            a_y = b_x(matrix(y_scaled[elems]), sy_mat, e_mu, p, kernel_type)
+            a_yp =  b_x(matrix(yp_scaled[elems]), syp_mat, e_mu, p, kernel_type)
 
             # part 1 of the sum
             aj = cumsum(a_y)
@@ -580,12 +544,22 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
 
           }else{
             # sy matrix
-            sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[i]))
-            syp_mat = solve(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h[j]))
+            if(check_inv(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[i]))[1]== TRUE){
+              sy_mat = solve(S_x(as.matrix(y_scaled[y_elems]), p, kernel_type)/(n*h[i]))
+            } else{
+              singular_flag = TRUE
+              sy_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+            }
+            if(check_inv(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h[j]))[1]== TRUE){
+              syp_mat = solve(S_x(as.matrix(yp_scaled[yp_elems]), p, kernel_type)/(n*h[j]))
+            } else{
+              singular_flag = TRUE
+              syp_mat = matrix(0L, nrow = length(e_mu), ncol = length(e_mu))
+            }
 
             # computing y and yprime vectors
-            a_y = b_x(y_scaled[elems], sy_mat, e_mu, p, kernel_type)
-            a_yp =  b_x(yp_scaled[elems], syp_mat, e_mu, p, kernel_type)
+            a_y = b_x(matrix(y_scaled[elems]), sy_mat, e_mu, p, kernel_type)
+            a_yp =  b_x(matrix(yp_scaled[elems]), syp_mat, e_mu, p, kernel_type)
 
             # part 1 of the sum
             aj = cumsum(a_y)
@@ -598,21 +572,28 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
             t_2 = (n-k) * a_yp[k] * aj[k]
             t_3 = (n-k) * a_y[k] * ak[k]
             t_4 = (n-k)^2 * a_y[k] * a_yp[k]
-            c_hat[i, j] = c_hat[i, j] + bx[elems[k]]^2 * (t_1 + t_2 + t_3 + t_4)
+            c_hat[i, j] = c_hat[i, j] + bx[1, elems[k]]^2 * (t_1 + t_2 + t_3 + t_4)
           }
         }
         # estimated means
-        theta_y = fhat(x_data=x_data, y_data=y_data, x=x, y_grid=y, p=p, q=q,
-                       mu=mu, nu=nu, h=h[i], kernel_type=kernel_type)$est
-        theta_yp = fhat(x_data=x_data, y_data=y_data, x=x, y_grid=y_prime, p=p,
-                        q=q, mu=mu, nu=nu, h=h[j], kernel_type=kernel_type)$est
+        if (mu==0){
+          theta_y = fhat(x_data=x_data, y_data=y_data, x=x, y_grid=y, p=2,
+                         q=1, mu=0, nu=0, h=h[i], kernel_type=kernel_type)$est
+          theta_yp = fhat(x_data=x_data, y_data=y_data, x=x, y_grid=y_prime,
+                          p=2, q=1, mu=0, nu=0, h=h[j], kernel_type=kernel_type)$est
+
+        }else{
+          theta_y = fhat(x_data=x_data, y_data=y_data, x=x, y_grid=y, p=p,
+                         q=q, mu=mu, nu=nu, h=h[i], kernel_type=kernel_type)$est
+          theta_yp = fhat(x_data=x_data, y_data=y_data, x=x, y_grid=y_prime,
+                          p=p, q=q, mu=mu, nu=nu, h=h[j], kernel_type=kernel_type)$est
+        }
 
         # filling matrix, using symmetry
-        c_hat[i, j] = c_hat[i, j]/(n*(n-1)^2) - theta_y*theta_yp/n
+        c_hat[i, j] = c_hat[i, j]/(n*(n-1)^2) - theta_y*theta_yp/n^2
         c_hat[j, i] = c_hat[i, j]
       }
     }
-    # c_hat = c_hat/(n^2*h^(2*d+2*mu+2*nu+2))
 
     if(mu==0){
       c_hat = sweep(sweep(c_hat, MARGIN=1, FUN="*", STATS=1/(n*h^(d+mu+nu))),
@@ -624,7 +605,7 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
     }
   }
 
-  return(c_hat)
+  return(list("cov" = c_hat, "singular_flag" = singular_flag))
 }
 #######################################################################################
 # Supplemental Functions
@@ -639,35 +620,10 @@ cov_hat = function(x_data, y_data, x, y_grid, p, q, mu, nu, h, kernel_type){
 #' @return vector of products for each data point.
 #' @keywords internal
 b_x = function(datavec, s_mat, e_vec, q, kernel_type){
-  eff_n = length(datavec)
+  eff_n = length(datavec[,1])
   Rq = matrix(0L, ncol = eff_n)
   for (i in 1:eff_n){
-    Rq[i] = (poly_base(datavec[i], q) * kernel_eval(datavec[i], kernel_type)) %*% (s_mat %*% e_vec)
+    Rq[i] = (poly_base(datavec[i, ], q) * kernel_eval(datavec[i, ], kernel_type)) %*% (s_mat %*% e_vec)
   }
   return(Rq)
 }
-
-# bias_corr = function(y_data, x_data, p, q, mu, nu, h, kernel_type){
-#   y_data= as.matrix(y_data)
-#   x_data = as.matrix(x_data)
-#   n = length(y_data)
-#   d=1
-#   # x basis vector
-#   e_nu = basis_vec(x, q, nu)
-#
-#   # y basis vector
-#   e_mu = basis_vec(1, p, mu)
-#   sy_mat = solve(S_x(y_data, p, kernel_type)/(n*h))
-#   sx_mat = solve(S_x(x_data, q, kernel_type)/(n*h^d))
-#   u = matrix(0L, nrow=n)
-#   for(i in 1:n){
-#     y_i = y_data[i]
-#     x_i = x_data[i]
-#     for (j in 1:n){
-#       y_j = y_data[j]
-#       if(j!=i){
-#         u[i] = uij(y_i, y_j, x_i, y, x, kernel_type, symat, sxmat, p, q, mu, nu)
-#       }
-#     }
-#   }
-# }
